# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import datetime

from logging.handlers import RotatingFileHandler

RED = "\033[91m"
BOLD = "\033[1m"
WARNING = "\033[91m"
ENDC = "\033[0m"


def init_logging(strategy, quiet_mode, name, args):
    fn = (
        datetime.datetime.now().strftime(strategy + "-%Y-%m-%d-%M:%S-%p")
        + "-"
        + str(name)
        + ".log"
    )
    log_fn = args.logfolder + "/" + fn
    logging.basicConfig(
        handlers=[
            RotatingFileHandler(filename=log_fn,
                                maxBytes=1024 * 1024, backupCount=5)
        ],
        format="%(asctime)s %(message)s",
        datefmt="[%Y/%m/%d %I:%M:%S %p]",
        level=logging.DEBUG,
    )

    if not quiet_mode:
        console = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s %(message)s", datefmt="[%Y/%m/%d %I:%M:%S %p]"
        )
        console.setLevel(logging.INFO)
        console.setFormatter(formatter)
        logging.getLogger().addHandler(console)


def log_strategy_num_seeds(strategy, seeds, targets):
    num_targets = len(targets)
    num_seeds = len(seeds)
    logging.info(
        "Strategy: "
        + strategy
        + ", "
        + str(num_targets)
        + " testing targets, "
        + str(num_seeds)
        + " seeds"
    )


def log_generation_attempt(args):
    logging.debug(
        "Attempting to generate " + str(args.iterations) + " mutants"
    )


def log_finished_generations(args, unsuccessful):
    successful = args.iterations - unsuccessful
    logging.debug(
        "Finished generations: "
        + str(successful)
        + " successful, "
        + str(unsuccessful)
        + " unsuccessful"
    )


def log_crash_trigger(path):
    logging.debug("Crash! Stop testing on this seed.")
    logging.info(BOLD + WARNING + "Detected crash bug: " + path + ENDC)


def log_ignore_list_mutant(solver_cli):
    logging.debug("Invalid mutant:ignore_list. solver=" + str(solver_cli))


def log_duplicate_trigger():
    logging.debug("Duplicate. Stop testing on this seed.")


def log_segfault_trigger(args, path, i):
    logging.debug(
        str(i) + "/" + str(args.iterations)
        + " Segfault! Stop testing on this seed."
    )
    logging.info(BOLD + WARNING + "Detected segfault: " + path + ENDC)


def log_solver_timeout(args, solver_cli, i):
    logging.debug(
        str(i)
        + "/"
        + str(args.iterations)
        + " Solver timeout occurred. sol="
        + str(solver_cli)
    )


def log_soundness_trigger(args, i, path):
    logging.debug(
        str(i)
        + "/"
        + str(args.iterations)
        + " Soundness bug! Stop testing on this seed."
    )
    logging.info(BOLD + WARNING + "Detected soundness bug! " + path + ENDC)


def log_invalid_mutant(args, i):
    logging.debug(
        str(i)
        + "/"
        + str(args.iterations)
        + " Invalid mutant:no '^sat$' or '^unsat$' in output."
    )



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Generated from SMTLIBv2.g4 by ANTLR 4.8
from antlr4 import *
from yinyang.src.parsing.Ast import *
from yinyang.src.parsing.SMTLIBv2Parser import SMTLIBv2Parser

# This class defines a complete generic visitor for a parse tree produced by SMTLIBv2Parser.
class SMTLIBv2Visitor(ParseTreeVisitor):
    # Visit a parse tree produced by SMTLIBv2Parser#response.
    def visitResponse(self, ctx: SMTLIBv2Parser.ResponseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#generalReservedWord.
    def visitGeneralReservedWord(self, ctx: SMTLIBv2Parser.GeneralReservedWordContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#simpleSymbol.
    def visitSimpleSymbol(self, ctx: SMTLIBv2Parser.SimpleSymbolContext):
        return ctx.getText()

    # Visit a parse tree produced by SMTLIBv2Parser#quotedSymbol.
    def visitQuotedSymbol(self, ctx: SMTLIBv2Parser.QuotedSymbolContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#predefSymbol.
    def visitPredefSymbol(self, ctx: SMTLIBv2Parser.PredefSymbolContext):
        return ctx.getText()

    # Visit a parse tree produced by SMTLIBv2Parser#predefKeyword.
    def visitPredefKeyword(self, ctx: SMTLIBv2Parser.PredefKeywordContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#symbol.
    def visitSymbol(self, ctx: SMTLIBv2Parser.SymbolContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#numeral.
    def visitNumeral(self, ctx: SMTLIBv2Parser.NumeralContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#decimal.
    def visitDecimal(self, ctx: SMTLIBv2Parser.DecimalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#hexadecimal.
    def visitHexadecimal(self, ctx: SMTLIBv2Parser.HexadecimalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#binary.
    def visitBinary(self, ctx: SMTLIBv2Parser.BinaryContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#string.
    def visitString(self, ctx: SMTLIBv2Parser.StringContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#keyword.
    def visitKeyword(self, ctx: SMTLIBv2Parser.KeywordContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#spec_constant.
    def visitSpec_constant(self, ctx: SMTLIBv2Parser.Spec_constantContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#s_expr.
    def visitS_expr(self, ctx: SMTLIBv2Parser.S_exprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#index.
    def visitIndex(self, ctx: SMTLIBv2Parser.IndexContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#identifier.
    def visitIdentifier(self, ctx: SMTLIBv2Parser.IdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#attribute_value.
    def visitAttribute_value(self, ctx: SMTLIBv2Parser.Attribute_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#attribute.
    def visitAttribute(self, ctx: SMTLIBv2Parser.AttributeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#qual_identifier.
    def visitQual_identifier(self, ctx: SMTLIBv2Parser.Qual_identifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#var_binding.
    def visitVar_binding(self, ctx: SMTLIBv2Parser.Var_bindingContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#sorted_var.
    def visitSorted_var(self, ctx: SMTLIBv2Parser.Sorted_varContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#pattern.
    def visitPattern(self, ctx: SMTLIBv2Parser.PatternContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#match_case.
    def visitMatch_case(self, ctx: SMTLIBv2Parser.Match_caseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#sort_symbol_decl.
    def visitSort_symbol_decl(self, ctx: SMTLIBv2Parser.Sort_symbol_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#meta_spec_constant.
    def visitMeta_spec_constant(self, ctx: SMTLIBv2Parser.Meta_spec_constantContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#fun_symbol_decl.
    def visitFun_symbol_decl(self, ctx: SMTLIBv2Parser.Fun_symbol_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#par_fun_symbol_decl.
    def visitPar_fun_symbol_decl(self, ctx: SMTLIBv2Parser.Par_fun_symbol_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#theory_attribute.
    def visitTheory_attribute(self, ctx: SMTLIBv2Parser.Theory_attributeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#theory_decl.
    def visitTheory_decl(self, ctx: SMTLIBv2Parser.Theory_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#logic_attribue.
    def visitLogic_attribue(self, ctx: SMTLIBv2Parser.Logic_attribueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#logic.
    def visitLogic(self, ctx: SMTLIBv2Parser.LogicContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#sort_dec.
    def visitSort_dec(self, ctx: SMTLIBv2Parser.Sort_decContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#selector_dec.
    def visitSelector_dec(self, ctx: SMTLIBv2Parser.Selector_decContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#constructor_dec.
    def visitConstructor_dec(self, ctx: SMTLIBv2Parser.Constructor_decContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#datatype_dec.
    def visitDatatype_dec(self, ctx: SMTLIBv2Parser.Datatype_decContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#function_dec.
    def visitFunction_dec(self, ctx: SMTLIBv2Parser.Function_decContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#function_def.
    def visitFunction_def(self, ctx: SMTLIBv2Parser.Function_defContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#prop_literal.
    def visitProp_literal(self, ctx: SMTLIBv2Parser.Prop_literalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_assert.
    def visitCmd_assert(self, ctx: SMTLIBv2Parser.Cmd_assertContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_checkSat.
    def visitCmd_checkSat(self, ctx: SMTLIBv2Parser.Cmd_checkSatContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_checkSatAssuming.
    def visitCmd_checkSatAssuming(
        self, ctx: SMTLIBv2Parser.Cmd_checkSatAssumingContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_declareConst.
    def visitCmd_declareConst(self, ctx: SMTLIBv2Parser.Cmd_declareConstContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_declareDatatype.
    def visitCmd_declareDatatype(self, ctx: SMTLIBv2Parser.Cmd_declareDatatypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_declareDatatypes.
    def visitCmd_declareDatatypes(
        self, ctx: SMTLIBv2Parser.Cmd_declareDatatypesContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_declareFun.
    def visitCmd_declareFun(self, ctx: SMTLIBv2Parser.Cmd_declareFunContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_declareSort.
    def visitCmd_declareSort(self, ctx: SMTLIBv2Parser.Cmd_declareSortContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_defineFun.
    def visitCmd_defineFun(self, ctx: SMTLIBv2Parser.Cmd_defineFunContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_defineFunRec.
    def visitCmd_defineFunRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunRecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_defineFunsRec.
    def visitCmd_defineFunsRec(self, ctx: SMTLIBv2Parser.Cmd_defineFunsRecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_defineSort.
    def visitCmd_defineSort(self, ctx: SMTLIBv2Parser.Cmd_defineSortContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_echo.
    def visitCmd_echo(self, ctx: SMTLIBv2Parser.Cmd_echoContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_exit.
    def visitCmd_exit(self, ctx: SMTLIBv2Parser.Cmd_exitContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getAssertions.
    def visitCmd_getAssertions(self, ctx: SMTLIBv2Parser.Cmd_getAssertionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getAssignment.
    def visitCmd_getAssignment(self, ctx: SMTLIBv2Parser.Cmd_getAssignmentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getInfo.
    def visitCmd_getInfo(self, ctx: SMTLIBv2Parser.Cmd_getInfoContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getModel.
    def visitCmd_getModel(self, ctx: SMTLIBv2Parser.Cmd_getModelContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getOption.
    def visitCmd_getOption(self, ctx: SMTLIBv2Parser.Cmd_getOptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getProof.
    def visitCmd_getProof(self, ctx: SMTLIBv2Parser.Cmd_getProofContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getUnsatAssumptions.
    def visitCmd_getUnsatAssumptions(
        self, ctx: SMTLIBv2Parser.Cmd_getUnsatAssumptionsContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getUnsatCore.
    def visitCmd_getUnsatCore(self, ctx: SMTLIBv2Parser.Cmd_getUnsatCoreContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_getValue.
    def visitCmd_getValue(self, ctx: SMTLIBv2Parser.Cmd_getValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_pop.
    def visitCmd_pop(self, ctx: SMTLIBv2Parser.Cmd_popContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_push.
    def visitCmd_push(self, ctx: SMTLIBv2Parser.Cmd_pushContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_reset.
    def visitCmd_reset(self, ctx: SMTLIBv2Parser.Cmd_resetContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_resetAssertions.
    def visitCmd_resetAssertions(self, ctx: SMTLIBv2Parser.Cmd_resetAssertionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_setInfo.
    def visitCmd_setInfo(self, ctx: SMTLIBv2Parser.Cmd_setInfoContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_setLogic.
    def visitCmd_setLogic(self, ctx: SMTLIBv2Parser.Cmd_setLogicContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#cmd_setOption.
    def visitCmd_setOption(self, ctx: SMTLIBv2Parser.Cmd_setOptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#b_value.
    def visitB_value(self, ctx: SMTLIBv2Parser.B_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#option.
    def visitOption(self, ctx: SMTLIBv2Parser.OptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#info_flag.
    def visitInfo_flag(self, ctx: SMTLIBv2Parser.Info_flagContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#error_behaviour.
    def visitError_behaviour(self, ctx: SMTLIBv2Parser.Error_behaviourContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#reason_unknown.
    def visitReason_unknown(self, ctx: SMTLIBv2Parser.Reason_unknownContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#model_response.
    def visitModel_response(self, ctx: SMTLIBv2Parser.Model_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#info_response.
    def visitInfo_response(self, ctx: SMTLIBv2Parser.Info_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#valuation_pair.
    def visitValuation_pair(self, ctx: SMTLIBv2Parser.Valuation_pairContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#t_valuation_pair.
    def visitT_valuation_pair(self, ctx: SMTLIBv2Parser.T_valuation_pairContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#check_sat_response.
    def visitCheck_sat_response(self, ctx: SMTLIBv2Parser.Check_sat_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#echo_response.
    def visitEcho_response(self, ctx: SMTLIBv2Parser.Echo_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_assertions_response.
    def visitGet_assertions_response(
        self, ctx: SMTLIBv2Parser.Get_assertions_responseContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_assignment_response.
    def visitGet_assignment_response(
        self, ctx: SMTLIBv2Parser.Get_assignment_responseContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_info_response.
    def visitGet_info_response(self, ctx: SMTLIBv2Parser.Get_info_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_model_response.
    def visitGet_model_response(self, ctx: SMTLIBv2Parser.Get_model_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_option_response.
    def visitGet_option_response(self, ctx: SMTLIBv2Parser.Get_option_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_proof_response.
    def visitGet_proof_response(self, ctx: SMTLIBv2Parser.Get_proof_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_unsat_assump_response.
    def visitGet_unsat_assump_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_assump_responseContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_unsat_core_response.
    def visitGet_unsat_core_response(
        self, ctx: SMTLIBv2Parser.Get_unsat_core_responseContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#get_value_response.
    def visitGet_value_response(self, ctx: SMTLIBv2Parser.Get_value_responseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#specific_success_response.
    def visitSpecific_success_response(
        self, ctx: SMTLIBv2Parser.Specific_success_responseContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by SMTLIBv2Parser#general_response.
    def visitGeneral_response(self, ctx: SMTLIBv2Parser.General_responseContext):
        return self.visitChildren(ctx)


del SMTLIBv2Parser



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

sys.setrecursionlimit(100000)

from yinyang.src.parsing.Types import (
    sort2type,
    # Types
    BOOLEAN_TYPE, REAL_TYPE, INTEGER_TYPE, ROUNDINGMODE_TYPE, STRING_TYPE,
    REGEXP_TYPE, UNKNOWN, ARRAY_TYPE, BITVECTOR_TYPE, FP_TYPE,
    # Operator lists
    CORE_OPS, NUMERICAL_OPS, REAL_OPS, INT_OPS, REAL_INTS, STRING_OPS,
    ARRAY_OPS, BV_OPS, FP_OPS,
    # Operators
    NOT, AND, OR, XOR, IMPLIES, ITE, EQ, UNARY_MINUS, MINUS, PLUS, MULTIPLY,
    GT, GTE, LT, LTE, DIV, MOD, ABS, REAL_DIV, TO_REAL, TO_INT, IS_INT,
    CONCAT, STRLEN, LEXORD, REFLEX_CLOS, STR_PREFIXOF, STR_SUFFIXOF,
    STR_CONTAINS, RE_NONE, RE_ALL, RE_ALLCHAR, STR_IN_RE, RE_COMP, RE_KLENE,
    RE_OPT, RE_PLUS, RE_DIFF, RE_CONCAT, RE_UNION, RE_INTER, STR_AT,
    STR_SUBSTR, STR_INDEXOF, STR_REPLACE, STR_REPLACE_ALL, STR_REPLACE_RE,
    STR_REPLACE_RE_ALL, STR_TO_CODE, STR_TO_INT, STR_TO_RE, STR_FROM_CODE,
    STR_FROM_INT, STR_IS_DIGIT, RE_RANGE, SELECT, STORE, BV_CONCAT, BVNOT,
    BVNEG, BVAND, BVOR, BVXOR, BVADD, BVSUB, BVMUL, BVUDIV, BVUREM, BVSHL,
    BV_EXTRACT, BV_ZERO_EXTEND, BV_SIGN_EXTEND, BVLSHR, BVASHR, BVSDIV, BVULT,
    BVULE, BVSLT, BVSGT, FP_ABS, FP_NEG, FP_ADD, FP_SUB, FP_MUL, FP_DIV,
    FP_SQRT, FP_REM, FP_ROUND_TO_INTEGRAL, FP_NORMAL, FP_ISSUBNORMAL,
    FP_IS_ZERO, FP_ISINFINITE, FP_ISNAN, FP_ISNEGATIVE, FP_ISPOSITIVE, FP_LEQ,
    FP_LT, FP_GEQ, FP_GT, FP_EQ, FP_MIN, FP_MAX, FP_FMA, TO_FP_UNSIGNED, TO_FP
)

from yinyang.src.parsing.Ast import Assert


class Context:
    def __init__(self, globals, locals):
        self.globals = globals
        self.locals = locals

    def add_to_globals(self, var, type):
        if isinstance(type, str):
            self.globals[var] = sort2type(type)
        else:
            self.globals[var] = type

    def add_to_locals(self, var, type):
        if isinstance(type, str):
            self.locals[var] = sort2type(type)
        else:
            self.locals[var] = type


class TypeCheckError(Exception):
    def __init__(self, expr, subterm=None, expected=None, actual=None):
        self.message = "Ill-typed expression \nterm: \t \t"\
                       + expr.__str__() + "\n"
        if isinstance(subterm, list):
            s = "[" + subterm[0].__str__() + " "
            for term in subterm[1:]:
                s += term.__str__() + " "
            s += "]"
            self.message += "faulty subterm:\t" + subterm.__str__() + "\n"
        else:
            self.message += "faulty subterm:\t" + subterm.__str__() + "\n"
        self.message += "expected: \t" + str(expected) + "\n"
        self.message += "actual: \t" + str(actual)
        super().__init__(self.message)


class UnknownOperator(Exception):
    def __init__(self, op):
        self.message = "unknown function/constant " + op
        # sys.tracebacklimit = 0
        super().__init__(self.message)


def typecheck_not(expr, ctxt=[]):
    """(not Bool Bool)"""
    typ = typecheck_expr(expr.subterms[0], ctxt)
    if typ != BOOLEAN_TYPE:
        raise TypeCheckError(expr, expr.subterms[0], typ, BOOLEAN_TYPE)
    return BOOLEAN_TYPE


def typecheck_unary_minus(expr, ctxt=[]):
    """(- Int Int)
    (- Real Real)
    """
    typ = typecheck_expr(expr.subterms[0], ctxt)
    if typ not in [INTEGER_TYPE, REAL_TYPE]:
        raise TypeCheckError(
            expr, expr.subterms[0], [INTEGER_TYPE, REAL_TYPE], typ
        )
    return typecheck_expr(expr.subterms[0], ctxt)


def typecheck_nary_numeral_ret(expr, ctxt=[]):
    """(- Int Int Int :left-assoc)
    (+ Int Int Int :left-assoc)
    (* Int Int Int :left-assoc)
    (- Real Real Real :left-assoc)
    (+ Real Real Real :left-assoc)
    (* Real Real Real :left-assoc)
    (/ Real Real Real :left-assoc)
    """
    typ = typecheck_expr(expr.subterms[0], ctxt)
    if typ not in [INTEGER_TYPE, REAL_TYPE]:
        raise TypeCheckError(
            expr, expr.subterms[0], [INTEGER_TYPE, REAL_TYPE], typ
        )
    for term in expr.subterms[1:]:
        t = typecheck_expr(term, ctxt)
        if t not in [INTEGER_TYPE, REAL_TYPE]:
            raise TypeCheckError(expr, term, typ, t)
    return typ


def typecheck_nary_int_ret(expr, ctxt=[]):
    """(div Int Int Int :left-assoc)
    (mod Int Int Int)
    (abs Int Int)
    """
    for term in expr.subterms:
        t = typecheck_expr(term, ctxt)
        if t != INTEGER_TYPE:
            raise TypeCheckError(expr, term, INTEGER_TYPE, t)
    return INTEGER_TYPE


def is_subtype(t, tprime):
    if t == INTEGER_TYPE and tprime == REAL_TYPE:
        return True
    if isinstance(t, BITVECTOR_TYPE) and isinstance(tprime, BITVECTOR_TYPE):
        return True
    return False


def typecheck_eq(expr, ctxt=[]):
    """(par (A) (= A A Bool :chainable))
    (par (A) (distinct A A Bool :pairwise))
    """
    typ = typecheck_expr(expr.subterms[0], ctxt)
    for term in expr.subterms[1:]:
        t = typecheck_expr(term, ctxt)
        if t != typ:
            if not (is_subtype(t, typ) or is_subtype(typ, t)):
                raise TypeCheckError(expr, term, typ, t)
    return BOOLEAN_TYPE


def typecheck_ite(expr, ctxt=[]):
    """(par (A) (ite Bool A A A))"""
    typ = typecheck_expr(expr.subterms[0], ctxt)
    if typecheck_expr(expr.subterms[0], ctxt) != BOOLEAN_TYPE:
        t = typecheck_expr(expr.subterms[1], ctxt)
        if not (is_subtype(t, typ) or is_subtype(typ, t)):
            raise TypeCheckError(expr, expr.subterms[0], typ, BOOLEAN_TYPE)
    t1 = typecheck_expr(expr.subterms[1], ctxt)
    t2 = typecheck_expr(expr.subterms[2], ctxt)
    if t1 != t2:
        if not (is_subtype(t1, t2) or is_subtype(t2, t1)):
            raise TypeCheckError(expr, expr.subterms[2], t1, t2)
    return typecheck_expr(expr.subterms[1], ctxt)


def typecheck_nary_bool(expr, ctxt=[]):
    """
    (and Bool Bool Bool :left-assoc)
    (or Bool Bool Bool :left-assoc)
    (xor Bool Bool Bool :left-assoc)
    """
    for term in expr.subterms:
        typ = typecheck_expr(term, ctxt)
        if typ != BOOLEAN_TYPE:
            raise TypeCheckError(expr, term, BOOLEAN_TYPE, typ)
    return BOOLEAN_TYPE


def typecheck_comp_ops(expr, ctxt):
    """
    (<= Int Int Bool :chainable)
    (<  Int Int Bool :chainable)
    (>= Int Int Bool :chainable)
    (>  Int Int Bool :chainable)
    (<= Real Real Bool :chainable)
    (<  Real Real Bool :chainable)
    (>= Real Real Bool :chainable)
    (>  Real Real Bool :chainable)
    """
    typ = typecheck_expr(expr.subterms[0], ctxt)
    if typ not in [INTEGER_TYPE, REAL_TYPE]:
        raise TypeCheckError(
            expr, expr.subterms[0], [INTEGER_TYPE, REAL_TYPE], typ
        )

    for term in expr.subterms[1:]:
        t = typecheck_expr(term, ctxt)
        if t not in [INTEGER_TYPE, REAL_TYPE]:
            raise TypeCheckError(expr, term, [INTEGER_TYPE, REAL_TYPE], t)
    return BOOLEAN_TYPE


def typecheck_to_real(expr, ctxt):
    """(to_real Int Real)"""
    t = typecheck_expr(expr, ctxt)
    if t != INTEGER_TYPE:
        raise TypeCheckError(expr, expr, INTEGER_TYPE, t)
    return REAL_TYPE


def typecheck_to_int(expr, ctxt):
    """(to_int Real Int)"""
    t = typecheck_expr(expr, ctxt)
    if t != REAL_TYPE or t != INTEGER_TYPE:
        raise TypeCheckError(expr, expr, REAL_TYPE, t)
    return INTEGER_TYPE


def typecheck_is_int(expr, ctxt):
    """(is_int Real Bool)"""
    t = typecheck_expr(expr, ctxt)
    if t != REAL_TYPE or t != INTEGER_TYPE:
        raise TypeCheckError(expr, expr, REAL_TYPE, t)
    return BOOLEAN_TYPE


def typecheck_real_div(expr, ctxt):
    """(/ Real Real Real :left-assoc)"""
    for term in expr.subterms:
        t = typecheck_expr(term, ctxt)
        if t not in [REAL_TYPE, INTEGER_TYPE]:
            raise TypeCheckError(expr, expr, REAL_TYPE, t)
    return REAL_TYPE


def typecheck_string_concat(expr, ctxt):
    """(str.++ String String String :left-assoc)"""
    for term in expr.subterms:
        t = typecheck_expr(term, ctxt)
        if t != STRING_TYPE:
            raise TypeCheckError(expr, expr, STRING_TYPE, t)
    return STRING_TYPE


def typecheck_strlen(expr, ctxt):
    """(str.len String Int)"""
    t = typecheck_expr(expr.subterms[0], ctxt)
    if t != STRING_TYPE:
        raise TypeCheckError(expr, expr, STRING_TYPE, t)
    return INTEGER_TYPE


def typecheck_nary_string_rt_bool(expr, ctxt):
    """(str.< String String Bool :chainable)
    (str.<= String String Bool :chainable)
    (str.prefixof String String Bool)
    (str.suffixof String String Bool)
    (str.contains String String Bool)
    """
    for term in expr.subterms:
        t = typecheck_expr(term, ctxt)
        if t != STRING_TYPE:
            raise TypeCheckError(expr, expr, STRING_TYPE, t)
    return BOOLEAN_TYPE


def typecheck_str_to_re(expr, ctxt):
    """(str.to_re String RegLan)"""
    arg = expr.subterms[0]
    t = typecheck_expr(arg, ctxt)
    if t != STRING_TYPE:
        raise TypeCheckError(expr, expr, STRING_TYPE, t)
    return REGEXP_TYPE


def typecheck_regex_consts(expr, ctxt):
    """
    (re.none RegLan)
    (re.all RegLan)
    (re.allchar RegLan)
    """
    return REGEXP_TYPE


def typecheck_str_in_re(expr, ctxt):
    """(str.in_re String RegLan Bool)"""
    s = typecheck_expr(expr.subterms[0], ctxt)
    t = typecheck_expr(expr.subterms[1], ctxt)

    if s != STRING_TYPE:
        raise TypeCheckError(expr, expr, STRING_TYPE, t)

    if t != REGEXP_TYPE:
        raise TypeCheckError(expr, expr, REGEXP_TYPE, t)
    return BOOLEAN_TYPE


def typecheck_regex_unary(expr, ctxt):
    """
    (re.* RegLan RegLan)
    (re.comp RegLan RegLan)
    (re.opt RegLan RegLan)
    (re.+ RegLan RegLan)
    """
    t = typecheck_expr(expr, ctxt)
    if t != REGEXP_TYPE:
        raise TypeCheckError(expr, expr, REGEXP_TYPE, t)
    return REGEXP_TYPE


def typecheck_regex_binary(expr, ctxt):
    """
    (re.diff RegLan RegLan RegLan :left-assoc)
    (re.++ RegLan RegLan RegLan :left-assoc)
    (re.union RegLan RegLan RegLan :left-assoc)
    (re.inter RegLan RegLan RegLan :left-assoc)
    """
    for term in expr.subterms:
        t = typecheck_expr(term, ctxt)
        if t != REGEXP_TYPE:
            raise TypeCheckError(expr, term, REGEXP_TYPE, t)
    return REGEXP_TYPE


def typecheck_str_at(expr, ctxt):
    """
    (str.at String Int String)
    """
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    if t1 != STRING_TYPE:
        raise TypeCheckError(expr, expr, STRING_TYPE, t1)
    if t2 != INTEGER_TYPE:
        raise TypeCheckError(expr, expr, INTEGER_TYPE, t2)
    return STRING_TYPE


def typecheck_substr(expr, ctxt):
    """
    (str.substr String Int Int String)
    """
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    t3 = typecheck_expr(expr.subterms[2], ctxt)
    if t1 != STRING_TYPE or t2 != INTEGER_TYPE or t3 != INTEGER_TYPE:
        raise TypeCheckError(
            expr, expr, [STRING_TYPE, INTEGER_TYPE, INTEGER_TYPE], [t1, t2, t3]
        )
    return STRING_TYPE


def typecheck_index_of(expr, ctxt):
    """(str.indexof String String Int Int)"""
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    t3 = typecheck_expr(expr.subterms[2], ctxt)
    if t1 != STRING_TYPE or t2 != STRING_TYPE or t3 != INTEGER_TYPE:
        raise TypeCheckError(
            expr, expr, [STRING_TYPE, STRING_TYPE, INTEGER_TYPE], [t1, t2, t3]
        )
    return INTEGER_TYPE


def typecheck_replace(expr, ctxt):
    """
    (str.replace String String String String)
    (str.replace_all String String String String)
    """
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    t3 = typecheck_expr(expr.subterms[2], ctxt)
    if t1 != STRING_TYPE or t2 != STRING_TYPE or t3 != STRING_TYPE:
        raise TypeCheckError(
            expr, expr, [STRING_TYPE, STRING_TYPE, STRING_TYPE], [t1, t2, t3]
        )
    return STRING_TYPE


def typecheck_replace_re(expr, ctxt):
    """
    (str.replace_re String RegLan String String)
    (str.replace_re_all String RegLan String String)
    """
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    t3 = typecheck_expr(expr.subterms[2], ctxt)
    if (
        typecheck_expr(expr.subterms[0], ctxt) != STRING_TYPE
        or typecheck_expr(expr.subterms[1], ctxt) != REGEXP_TYPE
        or typecheck_expr(expr.subterms[2], ctxt) != STRING_TYPE
    ):
        raise TypeCheckError(
            expr, expr, [STRING_TYPE, STRING_TYPE, STRING_TYPE], [t1, t2, t3]
        )
    return STRING_TYPE


def typecheck_re_range(expr, ctxt):
    """
    (re.range String String RegLan)
    """
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)

    if t1 != STRING_TYPE or t2 != STRING_TYPE:
        raise TypeCheckError(expr, expr, [STRING_TYPE, STRING_TYPE], [t1, t2])
    return REGEXP_TYPE


def typecheck_str_to_int(expr, ctxt):
    """
    (str.to_code String Int)
    (str.to_int String Int)
    """
    t = typecheck_expr(expr.subterms[0], ctxt)

    if t != STRING_TYPE:
        raise TypeCheckError(expr, expr, STRING_TYPE, t)
    return INTEGER_TYPE


def typecheck_is_digit(expr, ctxt):
    """
    (str.is_digit String Bool)
    """
    t = typecheck_expr(expr.subterms[0], ctxt)
    if t != STRING_TYPE:
        raise TypeCheckError(expr, expr, BOOLEAN_TYPE, t)
    return BOOLEAN_TYPE


def typecheck_int_to_string(expr, ctxt):
    """
    (str.from_code Int String)
    (str.from_int Int String)
    """
    t = typecheck_expr(expr.subterms[0], ctxt)
    if t != INTEGER_TYPE:
        raise TypeCheckError(expr, expr, INTEGER_TYPE, t)
    return STRING_TYPE


def typecheck_string_ops(expr, ctxt):
    if expr.op == CONCAT:
        return typecheck_string_concat(expr, ctxt)
    if expr.op == STRLEN:
        return typecheck_strlen(expr, ctxt)
    if expr.op in [
            LEXORD, REFLEX_CLOS, STR_PREFIXOF, STR_SUFFIXOF, STR_CONTAINS
    ]:
        return typecheck_nary_string_rt_bool(expr, ctxt)
    if expr.op in [RE_NONE, RE_ALL, RE_ALLCHAR]:
        return typecheck_regex_consts(expr, ctxt)
    if expr.op == STR_IN_RE:
        return typecheck_str_in_re(expr, ctxt)
    if expr.op in [RE_KLENE, RE_COMP, RE_OPT, RE_PLUS]:
        return typecheck_regex_binary(expr, ctxt)
    if expr.op in [RE_DIFF, RE_CONCAT, RE_UNION, RE_INTER]:
        return typecheck_regex_binary(expr, ctxt)
    if expr.op == STR_AT:
        return typecheck_str_at(expr, ctxt)
    if expr.op == STR_SUBSTR:
        return typecheck_substr(expr, ctxt)
    if expr.op == STR_INDEXOF:
        return typecheck_index_of(expr, ctxt)
    if expr.op in [STR_REPLACE, STR_REPLACE_ALL]:
        return typecheck_replace(expr, ctxt)
    if expr.op in [STR_REPLACE_RE, STR_REPLACE_RE_ALL]:
        return typecheck_replace(expr, ctxt)
    if expr.op in [STR_TO_CODE, STR_TO_INT]:
        return typecheck_str_to_int(expr, ctxt)
    if expr.op == STR_TO_RE:
        return typecheck_str_to_re(expr, ctxt)
    if expr.op in [STR_FROM_CODE, STR_FROM_INT]:
        return typecheck_int_to_string(expr, ctxt)
    if expr.op == STR_IS_DIGIT:
        return typecheck_is_digit(expr, ctxt)
    if expr.op == RE_RANGE:
        return typecheck_re_range(expr, ctxt)


def typecheck_select(expr, ctxt):
    """
    (select (Array X Y) X Y)
    """
    array_type = typecheck_expr(expr.subterms[0], ctxt)
    if isinstance(array_type, ARRAY_TYPE):
        raise TypeCheckError(expr, expr, ARRAY_TYPE, array_type)
    x_type = typecheck_expr(expr.subterms[1], ctxt)
    if x_type != array_type.index_type:
        raise TypeCheckError(expr, expr, array_type.index_type, x_type)
    return array_type.payload_type


def typecheck_store(expr, ctxt):
    """
    (store (Array X Y) X Y (Array X Y)))
    """
    array_type = typecheck_expr(expr.subterms[0], ctxt)
    if isinstance(array_type, ARRAY_TYPE):
        raise TypeCheckError(expr, expr, ARRAY_TYPE, array_type)
    x_type = typecheck_expr(expr.subterms[1], ctxt)
    y_type = typecheck_expr(expr.subterms[2], ctxt)
    if x_type != array_type.index_type and y_type != array_type.payload_type:
        raise TypeCheckError(
            expr,
            expr,
            [array_type.index_type, array_type.payload_type],
            [x_type, y_type],
        )
    return array_type


def typecheck_array_ops(expr, ctxt):
    if expr.op == SELECT:
        return typecheck_select(expr, ctxt)
    if expr.op == STORE:
        return typecheck_store(expr, ctxt)


def typecheck_bv_concat(expr, ctxt):
    """
    (concat (_ BitVec i) (_ BitVec j) (_ BitVec m))
    """
    arg1, arg2 = expr.subterms[0], expr.subterms[1]
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    if not isinstance(t1, BITVECTOR_TYPE) or\
       not isinstance(t2, BITVECTOR_TYPE):
        raise TypeCheckError(
            expr, [arg1, arg2], [BITVECTOR_TYPE, BITVECTOR_TYPE], [t1, t2]
        )
    bitwidth = t1.bitwidth + t2.bitwidth
    return BITVECTOR_TYPE(bitwidth)


def typecheck_bv_unary(expr, ctxt):
    """
    (op1 (_ BitVec m) (_ BitVec m))
    """
    arg = expr.subterms[0]
    t = typecheck_expr(expr.subterms[0], ctxt)
    if not isinstance(t, BITVECTOR_TYPE):
        raise TypeCheckError(expr, arg, BITVECTOR_TYPE, t)
    return t


def typecheck_bv_binary(expr, ctxt):
    """
    (op2 (_ BitVec m) (_ BitVec m) (_ BitVec m))
    """
    arg1, _ = expr.subterms[0], expr.subterms[1]
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    if not isinstance(t1, BITVECTOR_TYPE) or\
       not isinstance(t2, BITVECTOR_TYPE):
        expected = "[" + str(BITVECTOR_TYPE) + "," + str(BITVECTOR_TYPE) + "]"
        actual = "[" + str(t1) + "," + str(t2) + "]"
        raise TypeCheckError(expr, arg1, expected, actual)
    return BITVECTOR_TYPE(t1.bitwidth)


def typecheck_binary_bool_rt(expr, ctxt):
    """
    (bvult (_ BitVec m) (_ BitVec m) Bool)
    (bvule (_ BitVec m) (_ BitVec m) Bool)
    (bvslt (_ BitVec m) (_ BitVec m) Bool)
    """
    arg1, arg2 = expr.subterms[0], expr.subterms[1]
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)

    if not isinstance(arg1, BITVECTOR_TYPE) or\
       not isinstance(arg2, BITVECTOR_TYPE):
        raise TypeCheckError(
            expr, [arg1, arg2], [BITVECTOR_TYPE, BITVECTOR_TYPE], [t1, t2]
        )
    return BOOLEAN_TYPE


def typecheck_bv_ops(expr, ctxt):
    if expr.op == BV_CONCAT:
        return typecheck_bv_concat(expr, ctxt)
    if expr.op in [BVNOT, BVNEG]:
        return typecheck_bv_unary(expr, ctxt)
    if expr.op in [
        BVAND,
        BVOR,
        BVXOR,
        BVADD,
        BVSUB,
        BVMUL,
        BVUDIV,
        BVUREM,
        BVSHL,
        BVLSHR,
        BVASHR,
        BVSDIV,
    ]:
        return typecheck_bv_binary(expr, ctxt)
    if expr.op in [BVULT, BVULE, BVSLT, BVSGT]:
        return typecheck_binary_bool_rt(expr, ctxt)


def typecheck_fp_unary(expr, ctxt):
    """
    (fp.abs (_ FloatingPoint eb sb) (_ FloatingPoint eb sb))
    (fp.neg (_ FloatingPoint eb sb) (_ FloatingPoint eb sb))
    """
    arg = expr.subterms[0]
    t = typecheck_expr(expr.subterms[0], ctxt)
    if not isinstance(t, FP_TYPE):
        raise TypeCheckError(expr, arg, FP_TYPE, t)
    return typecheck_expr(expr.subterms[0], ctxt)


def typecheck_fp_binary_arith(expr, ctxt):
    """
     (fp.add RoundingMode (_ FloatingPoint eb sb) (_ FloatingPoint eb sb)
      (_ FloatingPoint eb sb))

    (fp.sub RoundingMode (_ FloatingPoint eb sb) (_ FloatingPoint eb sb)
      (_ FloatingPoint eb sb))

    (fp.mul RoundingMode (_ FloatingPoint eb sb) (_ FloatingPoint eb sb)
      (_ FloatingPoint eb sb))

    (fp.div RoundingMode (_ FloatingPoint eb sb) (_ FloatingPoint eb sb)
      (_ FloatingPoint eb sb))
    """
    arg1 = expr.subterms[0]
    arg2 = expr.subterms[1]
    arg3 = expr.subterms[2]
    t1 = typecheck_expr(arg1, ctxt)
    if t1 != ROUNDINGMODE_TYPE:
        raise TypeCheckError(expr, arg1, ROUNDINGMODE_TYPE, t1)

    t1 = typecheck_expr(arg2, ctxt)
    t2 = typecheck_expr(arg3, ctxt)
    if not isinstance(t1, FP_TYPE) or not isinstance(t2, FP_TYPE):
        raise TypeCheckError(expr, [arg2, arg3], [FP_TYPE, FP_TYPE])
    return typecheck_expr(arg2, ctxt)


def typecheck_fp_unary_bool_rt(expr, ctxt):
    """
    (fp.isNormal (_ FloatingPoint eb sb) Bool)
    (fp.isSubnormal (_ FloatingPoint eb sb) Bool)
    (fp.isZero (_ FloatingPoint eb sb) Bool)
    (fp.isInfinite (_ FloatingPoint eb sb) Bool)
    (fp.isNaN (_ FloatingPoint eb sb) Bool)
    (fp.isNegative (_ FloatingPoint eb sb) Bool)
    (fp.isPositive (_ FloatingPoint eb sb) Bool)
    """
    arg = expr.subterms[0]
    typ = typecheck_expr(arg, ctxt)
    if not isinstance(typ, FP_TYPE):
        raise TypeCheckError(expr, arg, FP_TYPE, typ)
    return BOOLEAN_TYPE


def typecheck_fp_comparison(expr, ctxt):
    """
    (fp.leq (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) Bool :chainable)
    (fp.lt  (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) Bool :chainable)
    (fp.geq (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) Bool :chainable)
    (fp.gt  (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) Bool :chainable)
    (fp.eq (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) Bool :chainable)
    """
    arg1 = expr.subterms[0]
    arg2 = expr.subterms[1]
    typ1 = typecheck_expr(arg1, ctxt)
    typ2 = typecheck_expr(arg2, ctxt)
    if not isinstance(typ1, FP_TYPE) or not isinstance(typ2, FP_TYPE):
        args = "[" + arg1.__str__() + "," + arg2.__str__() + "]"
        expected = "[" + str(FP_TYPE) + "," + str(FP_TYPE) + "]"
        actual = "[" + typ1.__str__() + "," + typ2.__str__() + "]"
        raise TypeCheckError(expr, args, expected, actual)
    return BOOLEAN_TYPE


def typecheck_fp_minmax(expr, ctxt):
    """
    (fp.min (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) (_ FloatingPoint eb sb)) 
    (fp.max (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) (_ FloatingPoint eb sb))
    """ # noqa E501
    arg1 = expr.subterms[0]
    arg2 = expr.subterms[1]
    if not isinstance(typecheck_expr(arg1, ctxt), FP_TYPE) or not isinstance(
        typecheck_expr(arg2, ctxt), FP_TYPE
    ):
        raise TypeCheckError(expr)
    return typecheck_expr(arg1, ctxt)


def typecheck_fp_fma(expr, ctxt):
    """
    (fp.fma RoundingMode (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) (_ FloatingPoint eb sb) (_ FloatingPoint eb sb))
    """ # noqa E501
    arg1 = expr.subterms[0]
    arg2 = expr.subterms[1]
    arg3 = expr.subterms[2]
    if (
        not isinstance(typecheck_expr(arg1, ctxt), FP_TYPE)
        or not isinstance(typecheck_expr(arg2, ctxt), FP_TYPE)
        or not isinstance(typecheck_expr(arg3, ctxt), FP_TYPE)
    ):
        raise TypeCheckError(expr)
    return typecheck_expr(arg1, ctxt)


def typecheck_fp_ops(expr, ctxt):
    if expr.op in [FP_ABS, FP_NEG]:
        return typecheck_fp_unary(expr, ctxt)
    if expr.op in [
        FP_ADD,
        FP_SUB,
        FP_MUL,
        FP_DIV,
        FP_SQRT,
        FP_REM,
        FP_ROUND_TO_INTEGRAL,
    ]:
        return typecheck_fp_binary_arith(expr, ctxt)
    if expr.op in [
        FP_NORMAL,
        FP_ISSUBNORMAL,
        FP_IS_ZERO,
        FP_ISINFINITE,
        FP_ISNAN,
        FP_ISNEGATIVE,
        FP_ISPOSITIVE,
    ]:
        return typecheck_fp_unary_bool_rt(expr, ctxt)
    if expr.op in [FP_LEQ, FP_LT, FP_GEQ, FP_GT, FP_EQ]:
        return typecheck_fp_comparison(expr, ctxt)
    if expr.op in [FP_MIN, FP_MAX]:
        return typecheck_fp_minmax(expr, ctxt)
    if expr.op == FP_FMA:
        return typecheck_fp_fma(expr, ctxt)


def typecheck_quantifiers(expr, ctxt):
    vars = expr.quantified_vars[0]
    types = expr.quantified_vars[1]
    for i in range(len(vars)):
        var, type = vars[i], types[i]
        ctxt.add_to_locals(var, type)
    t = typecheck_expr(expr.subterms[0], ctxt)
    if t != BOOLEAN_TYPE:
        raise TypeCheckError(expr, expr.subterms[0], BOOLEAN_TYPE, t)
    return BOOLEAN_TYPE


def typecheck_core(expr, ctxt):
    if expr.op == NOT:
        return typecheck_not(expr, ctxt)
    if expr.op in [AND, OR, XOR, IMPLIES]:
        return typecheck_nary_bool(expr, ctxt)
    if expr.op == ITE:
        return typecheck_ite(expr, ctxt)
    if expr.op in EQ:
        return typecheck_eq(expr, ctxt)


def typecheck_numeral(expr, ctxt):
    if expr.op == UNARY_MINUS and len(expr.subterms) == 1:
        return typecheck_unary_minus(expr, ctxt)
    if expr.op in [MINUS, PLUS, MULTIPLY]:
        return typecheck_nary_numeral_ret(expr, ctxt)
    if expr.op in [GT, GTE, LT, LTE]:
        return typecheck_comp_ops(expr, ctxt)


def typecheck_int_ops(expr, ctxt):
    if expr.op in [DIV, MOD, ABS]:
        return typecheck_nary_int_ret(expr, ctxt)


def typecheck_real_ops(expr, ctxt):
    if expr.op == REAL_DIV:
        return typecheck_real_div(expr, ctxt)


def typecheck_real_ints_ops(expr, ctxt):
    if expr.op == TO_REAL:
        return typecheck_to_real(expr, ctxt)
    if expr.op == TO_INT:
        return typecheck_to_int(expr, ctxt)
    if expr.op == IS_INT:
        return typecheck_is_int(expr, ctxt)


def typecheck_let_expression(expr, ctxt):
    n_var_binders = len(expr.var_binders)
    for i in range(n_var_binders):
        var = expr.var_binders[i]
        t = typecheck_expr(expr.let_terms[i], ctxt)
        ctxt.add_to_locals(var, t)
    return typecheck_expr(expr.subterms[0], ctxt)


def typecheck_label(expr, ctxt):
    return typecheck_expr(expr.subterms[0], ctxt)


def typecheck_to_fp_unsigned(expr, ctxt):
    """
    ((_ to_fp_unsigned eb sb) RoundingMode (_ BitVec m) (_ FloatingPoint eb sb))
    """  # noqa: 501
    eb, sb = int(expr.op.split(" ")[2]), int(expr.op.split(" ")[3].strip(")"))
    t1 = typecheck_expr(expr.subterms[0], ctxt)
    t2 = typecheck_expr(expr.subterms[1], ctxt)
    if not isinstance(t1, ROUNDINGMODE_TYPE)\
       or isinstance(t2, BITVECTOR_TYPE):
        raise TypeCheckError(expr)
    return FP_TYPE(eb, sb)


def typecheck_to_fp(expr, ctxt):
    """
    ((_ to_fp eb sb) (_ BitVec m) (_ FloatingPoint eb sb))
    ((_ to_fp eb sb) RoundingMode (_ FloatingPoint mb nb) (_ FloatingPoint eb sb))
    ((_ to_fp eb sb) RoundingMode Real (_ FloatingPoint eb sb))
    ((_ to_fp eb sb) RoundingMode (_ BitVec m) (_ FloatingPoint eb sb))
    """  # noqa: E501
    eb, sb = int(expr.op.split(" ")[2]), int(expr.op.split(" ")[3].strip(")"))
    if len(expr.subterms) == 1:
        bv = expr.subterms[0]
        t = typecheck_expr(expr.subterms[0], ctxt)
        if not isinstance(t, BITVECTOR_TYPE):
            raise TypeCheckError(expr, bv, BITVECTOR_TYPE, t)

    if len(expr.subterms) == 2:
        t1 = typecheck_expr(expr.subterms[0], ctxt)
        t2 = typecheck_expr(expr.subterms[1], ctxt)
        if (
            not (isinstance(t1, ROUNDINGMODE_TYPE) and isinstance(t2, FP_TYPE))
            or not (isinstance(t1, ROUNDINGMODE_TYPE) and t2 == REAL_TYPE)
            or not (
                isinstance(t1, ROUNDINGMODE_TYPE) and
                isinstance(t2, BITVECTOR_TYPE)
            )
        ):
            raise TypeCheckError(expr)

    return FP_TYPE(eb, sb)


def annotate(f, expr, ctxt):
    """
    f: function argument
    expr: expression
    ctxt: context
    :returns: type of expr
    """
    t = f(expr, ctxt)
    expr.type = t
    return t


def typecheck_expr(expr, ctxt=Context({}, {})):
    if expr.is_const:
        return expr.type
    if expr.is_var or expr.is_indexed_id:
        if expr.name in ctxt.locals:
            expr.type = ctxt.locals[expr.name]
            return ctxt.locals[expr.name]
        elif expr.name in ctxt.globals:
            expr.type = ctxt.globals[expr.name]
            return ctxt.globals[expr.name]
        return UNKNOWN
    elif expr.op:
        if expr.op in CORE_OPS:
            return annotate(typecheck_core, expr, ctxt)
        if expr.op in NUMERICAL_OPS:
            return annotate(typecheck_numeral, expr, ctxt)
        if expr.op in INT_OPS:
            return annotate(typecheck_int_ops, expr, ctxt)
        if expr.op in REAL_OPS:
            return annotate(typecheck_real_ops, expr, ctxt)
        if expr.op in REAL_INTS:
            return annotate(typecheck_real_ints_ops, expr, ctxt)
        if expr.op in STRING_OPS:
            return annotate(typecheck_string_ops, expr, ctxt)
        if expr.op in ARRAY_OPS:
            return annotate(typecheck_array_ops, expr, ctxt)
        if expr.op in FP_OPS:
            return annotate(typecheck_fp_ops, expr, ctxt)
        if expr.op in BV_OPS:
            return annotate(typecheck_bv_ops, expr, ctxt)

        # FP infix ops
        if TO_FP in expr.op:
            return annotate(typecheck_to_fp, expr, ctxt)

        if TO_FP_UNSIGNED in expr.op:
            return annotate(typecheck_to_fp_unsigned, expr, ctxt)

        # BV infix ops
        if (
            BV_EXTRACT in expr.op
            or BV_ZERO_EXTEND in expr.op
            or BV_SIGN_EXTEND in expr.op
        ):
            return annotate(typecheck_bv_unary, expr, ctxt)

        key = expr.op.__str__()
        if key in ctxt.globals:
            t = ctxt.globals[key].split(" ")[-1]
            expr.type = t
            return t
        raise UnknownOperator(expr.op)

    elif expr.quantifier:
        return annotate(typecheck_quantifiers, expr, ctxt)
    elif expr.let_terms:
        return annotate(typecheck_let_expression, expr, ctxt)
    elif expr.label:
        return annotate(typecheck_label, expr, ctxt)
    return UNKNOWN


def typecheck(formula, glob):
    """
    :formula: Script object representing formula
    :glob: glob variables for formula returned by parser
    :returns: context object ctxt
    """
    ctxt = Context(glob, {})
    for cmd in formula.commands:
        if isinstance(cmd, Assert):
            typecheck_expr(cmd.term, ctxt)
    return ctxt



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import string


def random_string(length=5):
    return "".join(random.sample(string.ascii_letters + string.digits, length))


def plain(cli):
    plain_cli = ""
    for token in cli.split(" "):
        plain_cli += token.split("/")[-1]
    return escape(plain_cli)


def escape(s):
    s = s.replace(".", "")
    s = s.replace("=", "")
    return s


def in_list(stdout, stderr, lst):
    stdstream = stdout + " " + stderr
    for err in lst:
        if err in stdstream:
            return True
    return False



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random

from yinyang.src.mutators.Mutator import Mutator


class TypeAwareOpMutation(Mutator):
    def __init__(self, formula, args):
        self.args = args
        self.formula = formula
        self.bidirectional = []
        self.unidirectional = []

        self.parse_config_file()

    def parse_config_file(self):
        with open(self.args.config) as f:
            lines = f.readlines()

        for line in lines:
            if ";" in line:
                continue
            if not line.strip():
                continue
            arity = -1
            if ":arity" in line:
                arity = line.split(":arity")[-1].split(" ")[-1].strip()
                line = " ".join(line.split(" ")[:-2])
            if "->" in line:
                op_from = line.split("->")[0].strip()
                op_to = line.split("->")[1].strip()
                self.unidirectional.append((arity, op_from, op_to))
                continue

            op_class = [op.strip() for op in line.split(",")]
            self.bidirectional.append((arity, op_class))

    def arities_mismatch(self, arity, op_occ):
        if arity == "2+" and len(op_occ.subterms) < 2:
            return True

        if arity == "1-" and len(op_occ.subterms) > 2:
            return True
        return False

    def get_replacee(self, op_occ):
        for b in self.bidirectional:
            arity, op_class = b[0], b[1]
            if self.arities_mismatch(arity, op_occ):
                continue

            if op_occ.op in op_class:
                diff = op_class.copy()
                diff.remove(op_occ.op)
                replacee = random.choice(diff)
                return replacee

            if op_occ.quantifier in op_class:
                diff = op_class.copy()
                diff.remove(op_occ.quantifier)
                replacee = random.choice(diff)
                return replacee

        for u in self.unidirectional:
            arity, op, replacee = u[0], u[1], u[2]
            if op_occ.op != op or op_occ.quantifier != op:
                continue
            if self.arities_mismatch(arity, op_occ):
                continue
            return replacee
        return None

    def mutate(self):
        success = False
        for _ in range(self.args.modulo):
            max_choices = len(self.formula.op_occs)
            for _ in range(max_choices):
                op_occ = random.choice(self.formula.op_occs)
                replacee = self.get_replacee(op_occ)
                if replacee:
                    success = True
                    op_occ.op = replacee
                    break
        return self.formula, success, False



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

BOOLEAN_TYPE = "Bool"
REAL_TYPE = "Real"
INTEGER_TYPE = "Int"
ROUNDINGMODE_TYPE = "RoundingMode"
STRING_TYPE = "String"
REGEXP_TYPE = "RegLan"
UNKNOWN = "Unknown"
ALL = "A"

TYPES = [
    BOOLEAN_TYPE,
    REAL_TYPE,
    INTEGER_TYPE,
    STRING_TYPE,
    REGEXP_TYPE,
    ROUNDINGMODE_TYPE,
]


def sort2type(sort):
    if "FloatingPoint" in sort:
        eb = int(sort.split(" ")[2])
        sb = int(sort.split(" ")[3][:-1])
        return FP_TYPE(eb, sb)

    if "BitVec" in sort:
        bitwith = int(sort.split(" ")[2][:-1])
        return BITVECTOR_TYPE(bitwith)
    return sort


class ARRAY_TYPE:
    def __init__(self, index_type, payload_type):
        self.index_type = index_type
        self.payload_type = payload_type

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.index_type == other.index_type
                and self.payload_type == other.payload_type
            )


class BITVECTOR_TYPE:
    def __init__(self, bitwidth):
        self.bitwidth = bitwidth

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.bitwidth == other.bitwidth

        if isinstance(other, str):
            return self.__str__() == other

    def __str__(self):
        #  (_ BitVec bitwidth)
        return "(_ BitVec " + str(self.bitwidth) + ")"


class FP_TYPE:
    def __init__(self, eb, sb):
        self.eb = eb
        self.sb = sb

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.eb == other.eb and self.sb == other.sb
        if isinstance(other, str):
            return self.__str__() == other

    def __str__(self):
        # (_ FloatingPoint eb sb)
        return "(_ FloatingPoint " + str(self.eb) + " " + str(self.sb) + ")"


# Core ops
NOT = "not"
AND = "and"
IMPLIES = "=>"
OR = "or"
XOR = "xor"
EQUAL = "="
DISTINCT = "distinct"
ITE = "ite"

CORE_OPS = [NOT, AND, IMPLIES, OR, XOR, EQUAL, DISTINCT, ITE]

# Numerical ops
UNARY_MINUS = "-"
MINUS = "-"
PLUS = "+"
MULTIPLY = "*"
ABS = "abs"
GTE = ">="
GT = ">"
LTE = "<="
LT = "<"

NUMERICAL_OPS = [UNARY_MINUS, MINUS, PLUS, MULTIPLY, ABS, GTE, GT, LTE, LT]


# specific Int ops
DIV = "div"
MOD = "mod"

INT_OPS = [DIV, MOD]

# specific real ops
REAL_DIV = "/"

REAL_OPS = [REAL_DIV]

# casting ops
TO_REAL = "to_real"
TO_INT = "to_int"
IS_INT = "is_int"

REAL_INTS = [REAL_DIV, TO_REAL, TO_INT, IS_INT]

EQ = [EQUAL, DISTINCT]
RT_BOOL = [NOT, AND, IMPLIES, OR, XOR]

# string ops
CONCAT = "str.++"
STRLEN = "str.len"
LEXORD = "str.<"
STR_TO_RE = "str.to_re"
STR_IN_RE = "str.in_re"
RE_NONE = "re.none"
RE_ALL = "re.all"
RE_ALLCHAR = "re.allchar"
RE_CONCAT = "re.++"
RE_UNION = "re.union"
RE_INTER = "re.inter"
RE_KLENE = "re.*"
REFLEX_CLOS = "str.<="
STR_AT = "str.at"
STR_SUBSTR = "str.substr"
STR_PREFIXOF = "str.prefixof"
STR_SUFFIXOF = "str.suffixof"
STR_CONTAINS = "str.contains"
STR_INDEXOF = "str.indexof"
STR_REPLACE = "str.replace"
STR_REPLACE_ALL = "str.replace_all"
STR_REPLACE_RE = "str.replace_re"
STR_REPLACE_RE_ALL = "str.replace_re_all"
RE_COMP = "re.comp"
RE_DIFF = "re.diff"
RE_PLUS = "re.+"
RE_OPT = "re.opt"
RE_RANGE = "re.range"
STR_IS_DIGIT = "str.is_digit"
STR_TO_CODE = "str.to_code"
STR_TO_INT = "str.to_int"
STR_FROM_CODE = "str.from_code"
STR_FROM_INT = "str.from_int"

STRING_OPS = [
    CONCAT,
    STRLEN,
    LEXORD,
    STR_TO_RE,
    STR_IN_RE,
    RE_NONE,
    RE_ALL,
    RE_ALLCHAR,
    RE_CONCAT,
    RE_UNION,
    RE_INTER,
    RE_KLENE,
    REFLEX_CLOS,
    STR_AT,
    STR_SUBSTR,
    STR_PREFIXOF,
    STR_SUFFIXOF,
    STR_CONTAINS,
    STR_INDEXOF,
    STR_REPLACE,
    STR_REPLACE_ALL,
    STR_REPLACE_RE,
    STR_REPLACE_RE_ALL,
    RE_COMP,
    RE_DIFF,
    RE_PLUS,
    RE_OPT,
    RE_RANGE,
    STR_IS_DIGIT,
    STR_TO_INT,
    STR_FROM_CODE,
    STR_FROM_INT,
]

# Array ops
SELECT = "select"
STORE = "store"
ARRAY_OPS = [SELECT, STORE]

# Bitvector ops
BV_CONCAT = "concat"
BVNOT = "bvnot"
BVNEG = "bvneg"
BVAND = "bvand"
BVOR = "bvor"
BVXOR = "bvxor"
BVADD = "bvadd"
BVSUB = "bvsub"
BVMUL = "bvmul"
BVUDIV = "bvudiv"
BVUREM = "bvurem"
BVSHL = "bvshl"
BVLSHR = "bvlshr"
BVASHR = "bvashr"
BVULT = "bvult"
BVULE = "bvule"
BVSLT = "bvslt"
BVSGT = "bvsgt"
BVSDIV = "bvsdiv"


BV_OPS = [
    BV_CONCAT,
    BVNOT,
    BVNEG,
    BVAND,
    BVOR,
    BVXOR,
    BVADD,
    BVSUB,
    BVMUL,
    BVUDIV,
    BVUREM,
    BVSHL,
    BVASHR,
    BVLSHR,
    BVULT,
    BVULE,
    BVSLT,
    BVSGT,
    BVSDIV,
]

"""
All function symbols with declaration of the form

  ((_ extract i j) (_ BitVec m) (_ BitVec n))

where
- i, j, m, n are numerals
- m > i >= j >= 0,
- n = i - j + 1
"""

BV_EXTRACT = "(_ extract"
BV_ZERO_EXTEND = "(_ zero_extend"
BV_SIGN_EXTEND = "(_ sign_extend"

# Floating Point ops
FP_ABS = "fp.abs"
FP_NEG = "fp.neg"
FP_ADD = "fp.add"
FP_SUB = "fp.sub"
FP_MUL = "fp.mul"
FP_DIV = "fp.div"
FP_SQRT = "fp.sqrt"
FP_REM = "fp.rem"
FP_ROUND_TO_INTEGRAL = "fp.roundToIntegral"
FP_FMA = "fp.fma"
FP_MIN = "fp.min"
FP_MAX = "fp.max"
FP_LEQ = "fp.leq"
FP_LT = "fp.lt"
FP_GEQ = "fp.geq"
FP_GT = "fp.gt"
FP_EQ = "fp.eq"
FP_NORMAL = "fp.isNormal"
FP_ISSUBNORMAL = "fp.isSubnormal"
FP_IS_ZERO = "fp.isZero"
FP_ISINFINITE = "fp.isInfinite"
FP_ISNAN = "fp.isNaN"
FP_ISNEGATIVE = "fp.isNegative"
FP_ISPOSITIVE = "fp.isPositive"

FP_OPS = [
    FP_ABS,
    FP_NEG,
    FP_ADD,
    FP_SUB,
    FP_MUL,
    FP_DIV,
    FP_SQRT,
    FP_REM,
    FP_ROUND_TO_INTEGRAL,
    FP_FMA,
    FP_MIN,
    FP_MAX,
    FP_LEQ,
    FP_LT,
    FP_GEQ,
    FP_GT,
    FP_EQ,
    FP_NORMAL,
    FP_ISSUBNORMAL,
    FP_IS_ZERO,
    FP_ISINFINITE,
    FP_ISNAN,
    FP_ISNEGATIVE,
    FP_ISPOSITIVE,
]

# FP infix ops
TO_FP = "to_fp"
TO_FP_UNSIGNED = "to_fp_unsigned"

# Quantifiers
EXISTS = "exists"
FORALL = "forall"



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import copy

from yinyang.src.parsing.Ast import Term
from yinyang.src.parsing.Types import (
    BOOLEAN_TYPE, REAL_TYPE, INTEGER_TYPE, ROUNDINGMODE_TYPE,
    STRING_TYPE, REGEXP_TYPE
)

type2num = {
    "Bool": 0,
    "Real": 1,
    "Int": 2,
    "RoundingMode": 3,
    "String": 4,
    "RegLan": 5,
    "Unknown": 6,
}


def get_subterms(expr):
    """
    Get all subexpression of term object expr.
    :returns: av_expr list of expressions
              expr_types list of types
              (s.t. expression e = av_expr[i] has type expr_types[i])
    """
    av_expr = []
    expr_types = []
    if isinstance(expr, Term):
        if expr.subterms:
            for s in expr.subterms:
                new_av, new_type = get_subterms(s)
                av_expr += new_av
                expr_types += new_type
            new_type = expr.type
            expr_types.append(new_type)
            av_expr.append(expr)
        else:
            av_expr.append(expr)
            expr_types.append(expr.type)
    else:
        if expr.term:
            new_av, new_type = get_subterms(expr.term)
            av_expr += new_av
            expr_types += new_type
    return av_expr, expr_types


def get_all_subterms(formula):
    """
    Get all expressions within a formula and their types.
    :returns: av_expr list of expressions
              expr_types list of types
              (s.t. expression e = av_expr[i] has type expr_types[i])
    """
    av_expr = []
    expr_type = []
    for i in range(len(formula.assert_cmd)):
        exps, typ = get_subterms(formula.assert_cmd[i])
        av_expr += exps
        expr_type += typ
    return av_expr, expr_type


def get_unique_subterms(formula):
    """
    Get all the unique expressions within a formula.
    :returns: unique_expr list of lists of unique expression for
     different types
    """
    av_expr, expr_type = get_all_subterms(formula)

    # Below index indicates what type of expressions are stored in each list
    # 0: Bool, 1: Real, 2: Int, 3: RoundingMode, 4: String, 5: Regex, 6: Ukn
    unique_expr = [[], [], [], [], [], []]
    for i in range(len(expr_type)):
        if expr_type[i] == BOOLEAN_TYPE:
            unique_expr[0].append(copy.deepcopy(av_expr[i]))
        elif expr_type[i] == REAL_TYPE:
            unique_expr[1].append(copy.deepcopy(av_expr[i]))
        elif expr_type[i] == INTEGER_TYPE:
            unique_expr[2].append(copy.deepcopy(av_expr[i]))
        elif expr_type[i] == ROUNDINGMODE_TYPE:
            unique_expr[3].append(copy.deepcopy(av_expr[i]))
        elif expr_type[i] == STRING_TYPE:
            unique_expr[4].append(copy.deepcopy(av_expr[i]))
        elif expr_type[i] == REGEXP_TYPE:
            unique_expr[5].append(copy.deepcopy(av_expr[i]))

    for i in range(6):
        if unique_expr[i]:
            temp = []
            temp.append(unique_expr[i][0])
            for j in range(1, len(unique_expr[i])):
                flag = 0
                for exp in temp:
                    if unique_expr[i][j] == exp:
                        flag = 1
                        pass
                if flag == 0:
                    temp.append(unique_expr[i][j])
            unique_expr[i] = temp
    return unique_expr


def local_defs(term, local):
    """
    term: term object
    local: list of local variables defined in the parent terms

    :returns: list of local vairables to be considered within the term
    """
    term = term.parent
    if term:
        if term.quantifier:
            for q_var in term.quantified_vars[0]:
                local.add(q_var)
        if term.let_terms:
            for var in term.var_binders:
                local.add(var)
        if term.parent:
            local_defs(term, local)
    return local


def local_compatible(t1, t2):
    """
    t1: term object
    t2: term object

    :returns: local compatibility of t2 for t, e.g. ,every local variables in
              t2 is defined in t1
    """
    loc_t1 = local_defs(t1, set())
    loc_t2 = local_defs(t2, set())
    return loc_t2.issubset(loc_t1)



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import copy
import string

from yinyang.src.parsing.Ast import Const, Var, Expr, Assert, DeclareConst


def gen_random_string(length):
    """
    Generate random string.
    """
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(length))


def get_first_assert_idx(template):
    """
    Find first assert's idx.
    """
    for first_ass_idx, cmd in enumerate(template.commands):
        if isinstance(cmd, Assert):
            return first_ass_idx
    return -1


def get_last_assert_idx(template):
    """
    Find last assert's idx.
    """
    last_idx = len(template.commands)
    for i, cmd in enumerate(template.commands):
        if isinstance(cmd, Assert):
            last_idx = i

    return last_idx


def get_constant_idx(template):
    """
    Checks whether template has a constant to be filled.
    :template:
    :returns: returns index of constant and -1 if there is no constant
    """
    for idx, decl in enumerate(template.commands):
        if not isinstance(decl, DeclareConst):
            break
        if decl.symbol == "c":
            return idx
    return -1


def get_constant_value(declare_const):
    """
    :declare_const: DeclareConst statement specifying the random constant
    :returns: expression for random constant
    """
    const_type = declare_const.sort

    if const_type == "Int":
        r = random.randint(-1000, 1000)
        if r < 0:
            return Expr(op="-",
                        subterms=[Const(name=str(-r), type=const_type)])
        else:
            return Const(name=str(r), type=const_type)

    if const_type == "Real":
        r = round(random.uniform(-1000, 1000), 5)
        if r < 0:
            return Expr(op="-",
                        subterms=[Const(name=str(-r), type=const_type)])
        else:
            return Const(str(r), type=const_type)

    if const_type == "Bool":
        return Const(random.choice(["true", "false"]), type=const_type)

    if const_type == "String":
        length = random.randint(0, 20)
        return Const('"' + gen_random_string(length) + '"', type=const_type)


def fill_template(x, y, template, var_type):
    """
    Binds the variable name to the template.
    :x: variable name of first formula
    :y: variable name of second formula
    :returns: bindded template
    """
    filled_template = copy.deepcopy(template)
    first_ass_idx = get_first_assert_idx(filled_template)
    z = Var(x + "_" + y + "_fused", var_type)

    # Detect whether template includes random variable c
    random_constant_idx = get_constant_idx(template)
    if random_constant_idx != -1:
        declare_const = template.commands[random_constant_idx]
        const_type = declare_const.sort
        const_expr = get_constant_value(declare_const)

        for ass in filled_template.commands[first_ass_idx:]:
            ass.term.substitute(Var("c", const_type), const_expr)

    # Bind occurrences x,y to template
    for ass in filled_template.commands[first_ass_idx:]:
        ass.term.substitute(Var("z", z.type), z)
        ass.term.substitute(Var("x", var_type), Var(x, var_type))
        ass.term.substitute(Var("y", var_type), Var(y, var_type))

    return filled_template


def inv_x(template):
    """
    :returns: inversion function term for variable occurrence x
    """
    if get_constant_idx(template) != -1:
        return template.commands[5].term.subterms[1]

    return template.commands[4].term.subterms[1]


def inv_y(template):
    """
    :returns: inversion function term for variable occurrence y
    """
    if get_constant_idx(template) != -1:
        return template.commands[6].term.subterms[1]

    return template.commands[5].term.subterms[1]


def fusion_contraints(template, var_type):
    """
    :returns: fusion constraints (i.e. last three asserts from filled template)
    """
    if var_type == "Int" or var_type == "Real":
        return template.commands[-3:]
    return template.commands[-3:-2]


def add_fusion_constraints(formula, asserts):
    """
    Add fusion constraint asserts to formula (after all other asserts)
    """
    last_ass_idx = get_last_assert_idx(formula)
    formula.commands = (
        formula.commands[: last_ass_idx + 1]
        + asserts
        + formula.commands[last_ass_idx + 1:]
    )


def add_var_decls(formula, declare_funs):
    """
    Add additional variable declarations to the formula (right before the first
    assert statement)
    """
    first_ass_idx = get_first_assert_idx(formula)
    formula.commands = (
        formula.commands[:first_ass_idx]
        + declare_funs
        + formula.commands[first_ass_idx:]
    )



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yinyang.src.base.Version import VERSION


header = (
    "yinyang -- a differential fuzzer for SMT solvers [version: "
    + VERSION
    + "]"
)

usage = """ yinyang [options] -o oracle solver_clis seed_file1 seed_file2 [optionally, more seed files]
       yinyang [options] -o oracle solver_clis seed_folder [optionally, more seed folders]

       solver_clis := "solver_cli1;solver_cli2;...;solver_clik"
       oracle      := {sat,unsat}
"""  # noqa: E501

short_description = """
yinyang is a tool for finding bugs in SMT solver by generating mutant SMT-LIB
scripts from a set of seed SMT-LIB scripts (.smt2 files) with known satisfiability
(i.e. either sat or unsat) and generates mutant formulas of the same satisfiability.
yinyang realizes semantic fusion and can work with a single SMT solver.

Example (basic usage):
$ yinyang -o sat "z3 model_validate=true" semantic-fusion-seeds/QF_NRA/sat
[2021/06/18 11:25:14 AM] Strategy: yinyang, 1 testing targets, 19228225 seeds
[2021/06/18 11:25:14 AM] Performed 0 solver calls (0.0 calls/s, eff: NaN, 0.0 mutants/s)
[2021/06/18 11:25:22 AM] Performed 2 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
[2021/06/18 11:25:30 AM] Performed 4 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
[2021/06/18 11:25:38 AM] Performed 6 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
[2021/06/18 11:25:47 AM] Performed 12 solver calls (0.4 calls/s, eff: 66.7%, 0.2 mutants/s)
[2021/06/18 11:25:55 AM] Performed 14 solver calls (0.3 calls/s, eff: 64.3%, 0.2 mutants/s)
User interrupt
16 seeds processed, 16 valid, 0 invalid
0 bug triggers found

This will by default randomly select smt2 files from `./semantic-fusion-seeds/QF_NRA/sat`
and generate 30 mutants per file. If a bug has been found, yinyang will store it
in `./bugs`. You can use the shortcut CTRL+C to terminate yinyang manually.
You can obtain pre-categorized seeds from the following GitHub repository:

    https://github.com/testsmt/semantic-fusion-seeds

For a listing of options and further information, use yinyang --help."""  # noqa: E501


long_description = """
yinyang is a tool for finding bugs in SMT solver by generating mutant SMT-LIB
scripts from a set of seed SMT-LIB scripts (.smt2 files) with known
satisfiability (i.e. either sat or unsat) and generates mutant formulas of the
same satisfiability. yinyang realizes semantic fusion which can work with a
single SMT solver. The idea behind semantic fusion is to fuse formula pairs of
the same satisfiability (both either sat or unsat) and so produce mutant
formulas of known satisfiability.

*** Example 1 (basic usage):

The following command runs yinyang on the the SMT solver z3 with pre-categorized
seed formulas [1] and shows its expected behavior.

    $ yinyang -o sat "z3 model_validate=true" semantic-fusion-seeds/QF_NRA/sat
    [2021/06/18 11:25:14 AM] Strategy: yinyang, 1 testing targets, 19228225 seeds
    [2021/06/18 11:25:14 AM] Performed 0 solver calls (0.0 calls/s, eff: NaN, 0.0 mutants/s)
    [2021/06/18 11:25:22 AM] Performed 2 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
    [2021/06/18 11:25:30 AM] Performed 4 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
    [2021/06/18 11:25:38 AM] Performed 6 solver calls (0.2 calls/s, eff: 50.0%, 0.1 mutants/s)
    [2021/06/18 11:25:47 AM] Performed 12 solver calls (0.4 calls/s, eff: 66.7%, 0.2 mutants/s)
    [2021/06/18 11:25:55 AM] Performed 14 solver calls (0.3 calls/s, eff: 64.3%, 0.2 mutants/s)
    User interrupt
    16 seeds processed, 16 valid, 0 invalid
    0 bug triggers found

This will by default randomly select smt2 files from `./semantic-fusion-seeds/QF_NRA/sat`
and generate 30 mutants per file. If a bug has been found, yinyang will store it
in `./bugs`.  You can terminate yinyang by CTLR + C. yinyang stores bug trigger in
./bugs and creates rolling logs up to 5MB per run in ./logs. Note, if you run yinyang
many times in a row, it might be worthwhile to disable logging to avoid an overflow
of the logs directory.

Every five seconds, yinyang will generate statistics on (1) solver calls,
(2) efficiency and (3) mutants per second. Efficiency corresponds to the percentage
of solver calls that did not timeout, cause parser or unsupported option errors.
It is expected that efficiency, solver calls and mutants per second can vary
quite a bit during a regular run of opfuzz. However, if the efficiency is
very low (e.g., < 30%), it may be a good idea to inspect the logs, see what
is going on and potentially adjust seed set and solver commandlines. Note, some
SMT solvers need options to support a particular set of seeds. E.g, cvc5 needs
the option "--strings-exp" to support string logic. Otherwise it will often
throw an error message.

[1] https://github.com/testsmt/semantic-fusion-seeds


"""  # noqa: E501


options = """
options:
    -o {sat,unsat}, --oracle {sat,unsat}
            oracle for fusion. Should match with the provided seed files
            (default: sat)
    -i <N>,  --iterations <N>
            iterations per seed (default: 30)
    -l <path>, --logfolder <path>
            log folder (default: ./logs)
    -t <secs>, --timeout <secs>
            timeout per SMT solver call in seconds (default: 8)
    -b <path>, --bugsfolder <path>
            set bug folder (default: ./bugs)
    -s <path>, --scratchfolder <path>
            temp folder to dump mutants. (default: ./scratch)
    -c <file>, --config <file>
            set custom fusion function file
    -L <bytes>, --limit <bytes>
            file size limit on seed formula in bytes (default: 100000)
    -n, --no-log    disable logging
    -q, --quiet     do not print statistics and other output
    -v, --version   show version number and exit
    -h, --help      show this help message and exit
"""



# MIT License
#
# Copyright (c) [2020 - 2021] The yinyang authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from yinyang.src.parsing.SMTLIBv2Parser import SMTLIBv2Parser
from yinyang.src.parsing.SMTLIBv2Visitor import SMTLIBv2Visitor
from yinyang.src.parsing.Ast import (
    Var,
    Const,
    Expr,
    Quantifier,
    LabeledTerm,
    LetBinding,
    Maximize,
    Minimize,
    Pop,
    Push,
    DeclareConst,
    DeclareFun,
    Define,
    DefineFun,
    DefineFunRec,
    DefineFunsRec,
    DefineConst,
    Script,
    AssertSoft,
    Assert,
    CheckSat,
    CheckSatAssuming,
    Simplify,
    Display,
    Eval,
    GetValue,
    FunDecl,
    SMTLIBCommand,
)
from yinyang.src.parsing.Types import (
    BITVECTOR_TYPE,
    INTEGER_TYPE,
    REAL_TYPE,
    STRING_TYPE,
    BOOLEAN_TYPE,
    REGEXP_TYPE,
    sort2type,
)


class AstException(Exception):
    pass


class AstVisitor(SMTLIBv2Visitor):
    def __init__(self, strict=True):
        self.strict = strict
        self.global_vars = {}
        self.sorts = {}

    def visitStart(self, ctx: SMTLIBv2Parser.StartContext):
        return self.visitScript(ctx.script())

    def visitScript(self, ctx: SMTLIBv2Parser.ScriptContext):
        cmds = []
        for c in ctx.command():
            cmds.append(self.visitCommand(c))
        return Script(cmds, self.global_vars)

    def add_to_globals(self, identifier, input_sorts, output_sort):
        if len(input_sorts) == 0:
            if output_sort in self.sorts:
                output_sort = self.sorts[output_sort]
            self.global_vars[identifier] = sort2type(output_sort)
        else:
            self.global_vars[identifier] =\
                sort2type(input_sorts + " " + output_sort)

    def handleCommand(self, ctx: SMTLIBv2Parser.CommandContext):
        if ctx.cmd_assert():
            return Assert(self.visitTerm(ctx.term()[0], {}))
        if ctx.cmd_assertSoft():
            attr = []
            for a in ctx.attribute():
                a = self.visitAttribute(a)
                attr.append(a)
            return AssertSoft(self.visitTerm(ctx.term()[0], {}), attr)
        if ctx.cmd_simplify():
            attr = []
            for a in ctx.attribute():
                a = self.visitAttribute(a)
                attr.append(a)
            return Simplify(self.visitTerm(ctx.term()[0], {}), attr)
        if ctx.cmd_minimize():
            return Minimize(self.visitTerm(ctx.term()[0], {}))
        if ctx.cmd_maximize():
            return Maximize(self.visitTerm(ctx.term()[0], {}))
        if ctx.cmd_display():
            return Display(self.visitTerm(ctx.term()[0], {}))
        if ctx.cmd_eval():
            return Eval(self.visitTerm(ctx.term()[0], {}))
        if ctx.cmd_declareConst():
            var = self.visitSymbol(ctx.symbol()[0])
            self.global_vars[var] = self.visitSort(ctx.sort()[0])
            decl = DeclareConst(
                self.visitSymbol(ctx.symbol()[0]),
                self.visitSort(ctx.sort()[0])
            )
            return decl
        if ctx.cmd_declareFun():
            input_sorts = []
            for sort in ctx.sort()[:-1]:
                input_sorts.append(self.visitSort(sort))
            output_sort = self.visitSort(ctx.sort()[-1])
            input_sorts = " ".join(input_sorts)
            identifier = self.visitSymbol(ctx.symbol()[0])
            self.add_to_globals(identifier, input_sorts, output_sort)
            return DeclareFun(identifier, input_sorts, output_sort)

        if ctx.cmd_define():
            return Define(
                self.visitSymbol(ctx.symbol()[0]),
                self.visitTerm(ctx.term()[0], {})
            )

        if ctx.cmd_defineConst():
            return DefineConst(
                self.visitSymbol(ctx.symbol()[0]),
                self.visitSort(ctx.sort()[0]),
                self.visitTerm(ctx.term()[0], {}),
            )

        if ctx.cmd_defineFun():
            sorted_vars = []
            for var in ctx.function_def().sorted_var():
                sorted_vars.append(self.visitSorted_var(var))
            identifier = self.visitSymbol(ctx.function_def().symbol())
            sorted_vars = " ".join(sorted_vars)
            self.add_to_globals(
                identifier, sorted_vars,
                self.visitSort(ctx.function_def().sort())
            )
            return DefineFun(
                identifier,
                sorted_vars,
                self.visitSort(ctx.function_def().sort()),
                self.visitTerm(ctx.function_def().term(), {}),
            )

        if ctx.cmd_defineFunRec():
            sorted_vars = []
            for var in ctx.function_def().sorted_var():
                sorted_vars.append(self.visitSorted_var(var))
            return DefineFunRec(
                self.visitSymbol(ctx.function_def().symbol()),
                sorted_vars,
                self.visitSort(ctx.function_def().sort()),
                self.visitTerm(ctx.function_def().term()),
                {},
            )

        if ctx.cmd_defineFunsRec():
            decls = []
            for decl in ctx.function_dec():
                decls.append(self.visitFunction_dec(decl))
            terms = []
            for term in ctx.term():
                terms.append(self.visitTerm(term, {}))
            return DefineFunsRec(decls, terms)

        if ctx.cmd_checkSat():
            terms = []
            for t in ctx.term():
                terms.append(self.visitTerm(t, {}))
            if len(terms) > 0:
                return CheckSat(terms)
            return CheckSat()

        if ctx.cmd_checkSatAssuming():
            terms = []
            for t in ctx.term():
                terms.append(self.visitTerm(t, {}))
            return CheckSatAssuming(terms)

        if ctx.cmd_getValue():
            terms = []
            for t in ctx.term():
                terms.append(self.visitTerm(t, {}))
            return GetValue(terms)

        if ctx.cmd_push():
            terms = []
            for t in ctx.term():
                terms.append(self.visitTerm(t, {}))
            if len(terms) > 0:
                return Push(terms)
            return Push()

        if ctx.cmd_pop():
            terms = []
            for t in ctx.term():
                terms.append(self.visitTerm(t, {}))
            if len(terms) > 0:
                return Pop(terms)
            return Pop()

    def visitFunction_dec(self, ctx: SMTLIBv2Parser.Function_decContext):
        sorted_vars = []
        for var in ctx.sorted_var():
            sorted_vars.append(self.visitSorted_var(var))

        return FunDecl(
            self.visitSymbol(
                ctx.symbol()), sorted_vars, self.visitSort(ctx.sort())
        )

    def visitSorted_var(self, ctx: SMTLIBv2Parser.Sorted_varContext):
        return (
            "("
            + self.visitSymbol(ctx.symbol())
            + " "
            + self.visitSort(ctx.sort())
            + ")"
        )

    def getString(self, ctx):
        start, stop = ctx.start.start, ctx.stop.stop
        return ctx.start.getInputStream().getText(start, stop)

    def visitCommand(self, ctx: SMTLIBv2Parser.CommandContext):
        if not self.strict:
            try:
                cmd = self.handleCommand(ctx)
                if not cmd:
                    return SMTLIBCommand(self.getString(ctx))
                else:
                    return cmd
            except Exception:
                return SMTLIBCommand(self.getString(ctx))
        else:
            cmd = self.handleCommand(ctx)
            if not cmd:

                return SMTLIBCommand(self.getString(ctx))
            else:
                return cmd

    def visitAttribute(self, ctx: SMTLIBv2Parser.AttributeContext):
        return (ctx.keyword().getText(), ctx.attribute_value().getText())

    def handle_quantifier(
            self,
            ctx: SMTLIBv2Parser.TermContext,
            quant,
            local_vars):

        subterms = []
        qvars = []
        qtypes = []
        for i in range(len(ctx.sorted_var())):
            qvar = self.visitSymbol(ctx.sorted_var()[i].symbol())
            qtype = self.visitSort(ctx.sorted_var()[i].sort())
            local_vars[qvar] = qtype
            qvars.append(qvar)
            qtypes.append(qtype)

        for t in ctx.term():
            subterms.append(self.visitTerm(t, local_vars))
        return Quantifier(quant, (qvars, qtypes), subterms)

    def visitSpec_constant(self, ctx: SMTLIBv2Parser.Spec_constantContext):
        """
        spec_constant
        : numeral
        | decimal
        | hexadecimal
        | binary
        | string
        | b_value
        | ParOpen GRW_Underscore ' bv' numeral numeral ParClose
        ;
        """
        if ctx.ParOpen():
            X, n = (
                ctx.numeral()[0].getText(),
                ctx.numeral()[1].getText().encode("utf-8").decode("utf-8")
            )
            return "(_ bv" + X + " " + n + ")", BITVECTOR_TYPE(int(n))
        if ctx.numeral():
            return ctx.getText().encode("utf-8").decode("utf-8"), INTEGER_TYPE
        if ctx.decimal():
            return ctx.getText().encode("utf-8").decode("utf-8"), REAL_TYPE
        if ctx.hexadecimal():
            return ctx.getText().encode("utf-8").decode("utf-8"), INTEGER_TYPE
        if ctx.binary():
            return ctx.getText().encode("utf-8").decode("utf-8"), INTEGER_TYPE
        if ctx.string():
            return ctx.getText().encode("utf-8").decode("utf-8"), STRING_TYPE
        if ctx.b_value():
            return ctx.getText().encode("utf-8").decode("utf-8"), BOOLEAN_TYPE
        if ctx.reg_const():
            return ctx.getText().encode("utf-8").decode("utf-8"), REGEXP_TYPE

    def visitTerm(self, ctx: SMTLIBv2Parser.TermContext, local_vars):
        """
        term
        : spec_constant
        | qual_identifier
        | ParOpen qual_identifier term+ ParClose
        | ParOpen GRW_Underscore ' bv' numeral numeral ParClose
        | ParOpen ParOpen GRW_Underscore qual_identifier term+ ParClose
          ParClose
        | ParOpen GRW_Let ParOpen var_binding+ ParClose term ParClose
        | ParOpen GRW_Forall ParOpen sorted_var+ ParClose term ParClose
        | ParOpen GRW_Exists ParOpen sorted_var+ ParClose term ParClose
        | ParOpen GRW_Match term ParOpen match_case+ ParClose ParClose
        | ParOpen GRW_Exclamation term attribute+ ParClose
        ;
        """

        if (
            ctx.ParOpen()
            and ctx.GRW_Exclamation()
            and ctx.term()
            and len(ctx.attribute()) >= 1
            and ctx.ParClose()
        ):
            term, label = self.visitTerm(ctx.term()[0]), self.visitAttribute(
                ctx.attribute()[0]
            )
            return LabeledTerm(label, [term])

        if (
            len(ctx.ParOpen()) == 2
            and ctx.GRW_Match()
            and ctx.term()
            and len(ctx.match_case()) >= 1
        ):
            raise AstException(
                "ParOpen GRW_Match term ParOpen match_case+ ParClose ParClose"
            )

        if len(ctx.ParOpen()) == 1 and ctx.GRW_Underscore() and ctx.numeral():
            bitwidth = ctx.symbol().getText().strip("bv")
            value = ctx.numeral().getText()
            return Const(name="(_ bv" + bitwidth + " " + value + ")")

        if len(ctx.ParOpen()) == 1 and ctx.GRW_Underscore() and ctx.numeral():
            bitwidth = ctx.symbol().getText().strip("bv")
            value = ctx.numeral().getText()
            return Const(name="(_ bv" + bitwidth + " " + value + ")")

        if (
            len(ctx.ParOpen()) == 2
            and ctx.GRW_Exists()
            and len(ctx.sorted_var()) >= 1
            and len(ctx.ParClose()) == 2
            and ctx.term()
        ):

            return self.handle_quantifier(ctx, "exists", local_vars)

        if (
            len(ctx.ParOpen()) == 2
            and ctx.GRW_Forall()
            and len(ctx.sorted_var()) >= 1
            and len(ctx.ParClose()) == 2
            and ctx.term()
        ):

            return self.handle_quantifier(ctx, "forall", local_vars)

        if (
            len(ctx.ParOpen()) == 2
            and ctx.GRW_Let()
            and ctx.var_binding()
            and len(ctx.ParClose()) == 2
            and ctx.term()
        ):

            terms = []
            var_list = []
            for b in ctx.var_binding():
                local_vars[self.visitSymbol(b.symbol())] = "Unknown"
                var_list.append(self.visitSymbol(b.symbol()))
                terms.append(self.visitTerm(b.term(), local_vars))
            subterms = []
            for sub in ctx.term():
                subterms.append(self.visitTerm(sub, local_vars=local_vars))
            return LetBinding(var_list, terms, subterms=subterms)

        if (
            ctx.ParOpen()
            and ctx.qual_identifier()
            and len(ctx.term()) >= 1
            and ctx.ParClose()
        ):

            op = self.visitQual_identifier(ctx.qual_identifier(), local_vars)
            subterms = []
            for term in ctx.term():
                subterms.append(self.visitTerm(term, local_vars))
            return Expr(op=op, subterms=subterms)

        if ctx.spec_constant():
            name, type = self.visitSpec_constant(ctx.spec_constant())
            return Const(name=name, type=type)

        if ctx.qual_identifier():
            return self.visitQual_identifier(ctx.qual_identifier(), local_vars)

        raise AstException("No match for term : ... |... |... ")

    def visitQual_identifier(
        self, ctx: SMTLIBv2Parser.Qual_identifierContext, local_vars
    ):
        """
        qual_identifier
        : identifier
        | ParOpen GRW_As identifier sort ParClose
        ;
        """

        if (
            ctx.ParOpen()
            and ctx.GRW_As()
            and ctx.identifier()
            and ctx.sort()
            and ctx.ParClose()
        ):
            raise AstException("ParOpen GRW_As identifier sort ParClose")

        if ctx.identifier():
            return self.visitIdentifier(ctx.identifier(), local_vars)

        raise AstException("No match for qual_identifier: ... |... |... ")

    def visitSimpleSymbol(self, ctx: SMTLIBv2Parser.SimpleSymbolContext):
        return ctx.getText()

    def visitQuotedSymbol(self, ctx: SMTLIBv2Parser.QuotedSymbolContext):
        return ctx.getText()

    def visitSymbol(self, ctx: SMTLIBv2Parser.SymbolContext):
        """
        symbol
        : simpleSymbol
        | quotedSymbol
        ;
        """

        if ctx.simpleSymbol():
            return self.visitSimpleSymbol(ctx.simpleSymbol())

        if ctx.quotedSymbol():
            return self.visitQuotedSymbol(ctx.quotedSymbol())

    def visitIdentifier(
            self,
            ctx: SMTLIBv2Parser.IdentifierContext,
            local_vars):
        """
        identifier
        : symbol
        | ParOpen GRW_Underscore symbol index+ ParClose
        ;
        """

        if (
            ctx.ParOpen()
            and ctx.GRW_Underscore()
            and ctx.symbol()
            and len(ctx.index()) >= 1
            and ctx.ParClose()
        ):
            symbol = self.visitSymbol(ctx.symbol())
            index = ctx.index()[0].getText()
            for ind in ctx.index()[1:]:
                index += " " + ind.getText()
            name = "(_ " + symbol + " " + index + ")"
            if name in local_vars:
                return Var(name=name, type=local_vars[name],
                           is_indexed_id=True)
            elif name in self.global_vars:
                return Var(name=name, type=self.global_vars[name],
                           is_indexed_id=True)
            else:
                return name

        if ctx.symbol():
            name = self.visitSymbol(ctx.symbol())
            if name in local_vars:
                return Var(name=name, type=local_vars[name])
            elif name in self.global_vars:
                return Var(name=name, type=self.global_vars[name])
            else:
                return self.visitSymbol(ctx.symbol())
        raise AstException("No match for identifier: ... |... |... ")

    def visitTerminal(self, ctx):
        return ctx.getText()

    def visitSort(self, ctx: SMTLIBv2Parser.SortContext):
        if len(ctx.sort()) >= 1:
            s = "(" + self.visitIdentifier(ctx.identifier(), {})
            for sort in ctx.sort():
                s += " " + self.visitSort(sort)
            return s + ")"
        return self.visitIdentifier(ctx.identifier(), {})
